# Storage Migration module

package PVE::Storage::StorageMigrate;

use strict;
use warnings;
use POSIX;
use Data::Dumper;

use PVE::Storage;
use PVE::QemuServer;
use PVE::Cluster;
use PVE::Tools qw(run_command);

sub disk_keys {
	my @names = [];
	
	@names = PVE::QemuServer::disknames();

	return @names;
}

sub vm_name {
	my ($volname) = @_;
	
	if ($volname =~ /^[a-zA-Z]+-(\d+)-.+/) {
		return $1;
	}
	elsif ($volname =~ /^(\d+)\/[a-zA-Z]+-\d+-.+/) {
		return $1;
	}
	else {
		die "$volname: Bad format\n";
	}
}

sub vol_format {
	my ($type, $volname) = @_;
	
	if ($volname =~ /.+\.(.+)$/) {
		if ($type eq 'dir' || $type eq 'nfs') {
			return $1;
		}
		if ($type eq 'lvm') {
			return $1 if ($1 eq 'raw');
			warn "Target [$type] only supports raw. Changing format to raw\n";
			return 'raw';
		}
	}
	return 'raw';
}

sub disk_name {
	my ($volname, $nosuffix) = @_;
	
	if ($volname =~ /^[a-zA-Z]+-\d+-(.+)/) {
		return $1 unless $nosuffix;
		my $path = $1;
		if ($path =~ /^(\w+-\d+)\./) {
			return $1;
		}
		return $path;
	}
	elsif ($volname =~ /^\d+\/[a-zA-Z]+-\d+-(.+)/) {
		return $1 unless $nosuffix;
		my $path = $1;
		if ($path =~ /^(\w+-\d+)\./) {
			return $1;
		}
		return $path;
	}
	else {
		die "$volname: Bad format\n";
	}
}

sub vol_name {
	my ($cfg, $volname) = @_;
	my ($vm, $disk);
	
	$vm = vm_name($volname);
	if ($cfg->{type} eq 'lvm') {
		$disk = disk_name($volname, 1);
		return "vm-$vm-$disk" if ($vm && $disk);
	}
	if ($cfg->{type} eq 'nfs' || $cfg->{type} eq 'dir') {
		$disk = disk_name($volname);
		return "$vm/vm-$vm-$disk" if ($vm && $disk);
	}
	die "$volname: Bad format\n";
}

sub volume_ha_enabled {
	my ($volname) = @_;
	
	my $vm = vm_name($volname);
	if ($vm) {
		my $cfg = PVE::Cluster::cfs_read_file('cluster.conf');
		my $vmid = PVE::Cluster::cluster_conf_lookup_pvevm($cfg, 0, $vm, 1);
		if ($vmid) {
			return 1;
		}
	}
	return 0;
}

my $migrate_info = sub {
	my ($cfg, $storeid, $volname, $target) = @_;
	my $info = {};
	
    if ($storeid && $target) {
		$info->{source}->{storeid} = $storeid;
		my $sinfo = PVE::Storage::storage_config($cfg, $storeid);
		foreach my $skey (keys %$sinfo) {
			$info->{source}->{$skey} = $sinfo->{$skey};
		}
		$info->{source}->{volname} = vol_name($info->{source}, $volname);
		$info->{source}->{plugin} = PVE::Storage::Plugin->lookup($sinfo->{type});
		$info->{source}->{vmid} = vm_name($volname);
		
		$info->{target}->{storeid} = $target;
		my $tinfo = PVE::Storage::storage_config($cfg, $target);
		foreach my $tkey (keys %$tinfo) {
			$info->{target}->{$tkey} = $tinfo->{$tkey};
		}
		if ($sinfo->{type} eq 'lvm' && ($tinfo->{type} eq 'dir' || $tinfo->{type} eq 'nfs')) {
			$info->{target}->{volname} = vol_name($info->{target}, $volname) . '.raw';
		} else {
			$info->{target}->{volname} = vol_name($info->{target}, $volname);
		}
		$info->{target}->{plugin} = PVE::Storage::Plugin->lookup($tinfo->{type});
		$info->{target}->{format} = vol_format($tinfo->{type}, $volname);
		$info->{target}->{vmid} = vm_name($volname);
		$info->{source}->{plugin}->activate_volume($storeid, $sinfo, $volname);
		my $bytes = $info->{source}->{plugin}->volume_size_info($sinfo, $storeid, $volname);
		$info->{target}->{size} = int($bytes / 1024);
		$info->{source}->{plugin}->deactivate_volume($storeid, $sinfo, $volname);
    }
    return $info;
};

my $lock_source = sub {
	my ($source) = @_;
	my ($cmd, $vol);
	
	if ($source->{type} eq 'lvm') {
		$vol = $source->{plugin}->path($source, $source->{volname}, $source->{storeid});
		$cmd = ['/sbin/lvchange', '-pr', "$vol"];
	} elsif ($source->{type} eq 'dir') {
		$vol = $source->{plugin}->path($source, $source->{volname}, $source->{storeid});
		$cmd = ['/usr/bin/chattr',  '+i', "$vol"];
	} elsif ($source->{type} eq 'nfs') {
		$vol = $source->{plugin}->path($source, $source->{volname}, $source->{storeid});
		$cmd = ['/bin/chmod',  'a=+r', "$vol"];
	}
	else {
		die $source->{type} . ": Locking not supported.\n";
	}

	run_command($cmd, errmsg => "lock '$vol' error");	
};

my $unlock_source = sub {
	my ($source) = @_;
	my ($cmd, $vol);
	
	if ($source->{type} eq 'lvm') {
		$vol = $source->{plugin}->path($source, $source->{volname}, $source->{storeid});
		$cmd = ['/sbin/lvchange', '-prw', "$vol"];
	} elsif ($source->{type} eq 'dir') {
		$vol = $source->{plugin}->path($source, $source->{volname}, $source->{storeid});
		$cmd = ['/usr/bin/chattr',  '-i', "$vol"];
	} elsif ($source->{type} eq 'nfs') {
		$vol = $source->{plugin}->path($source, $source->{volname}, $source->{storeid});
		$cmd = ['/bin/chmod',  'u+w', "$vol"];
	}
	else {
		die $source->{type} . ": Locking not supported.\n";
	}

	run_command($cmd, errmsg => "unlock '$vol' error");	
};

my $create_target = sub {
	my ($info) = @_;
	my $volname;
	
	my $target = $info->{target};
	my $plugin = $target->{plugin};
	
	$lock_source->($info->{source});

	if ($target->{volname} =~ /^\d+\/(.+)/) {
		$volname = $1;
	} else {
		$volname = $target->{volname};
	}
	$plugin->alloc_image($target->{storeid}, $target, $target->{vmid}, $target->{format}, $volname, $target->{size});
};

my $migrate_file_file = sub {
	my ($info) = @_;
	
	my $from_plugin = $info->{source}->{plugin};
	my $to_plugin = $info->{target}->{plugin};

	my $from = $from_plugin->path($info->{source}, $info->{source}->{volname}, $info->{source}->{storeid});
	my $to = $to_plugin->path($info->{target}, $info->{target}->{volname}, $info->{target}->{storeid});
    my $cmd = ['/usr/bin/qemu-img', 'convert', '-p', '-t', 'writeback', '-O', "$info->{target}->{format}", "$from", "$to"];

	#print Dumper($cmd);
    run_command($cmd, errmsg => "migrate '$from->$to' error");	
};

my $migrate_file_device = sub {
	my ($info) = @_;

	my $from_plugin = $info->{source}->{plugin};
	my $to_plugin = $info->{target}->{plugin};

	my $from = $from_plugin->path($info->{source}, $info->{source}->{volname}, $info->{source}->{storeid});
	my $to = $to_plugin->path($info->{target}, $info->{target}->{volname}, $info->{target}->{storeid});
    my $cmd = ['/usr/bin/qemu-img', 'convert', '-p', '-t', 'writeback', '-O', "$info->{target}->{format}", "$from", "$to"];

	#print Dumper($cmd);
    run_command($cmd, errmsg => "migrate '$from->$to' error");	
};

my $migrate_device_file = sub {
	my ($info) = @_;
	
	my $from_plugin = $info->{source}->{plugin};
	my $to_plugin = $info->{target}->{plugin};

	my $from = $from_plugin->path($info->{source}, $info->{source}->{volname}, $info->{source}->{storeid});
	my $to = $to_plugin->path($info->{target}, $info->{target}->{volname}, $info->{target}->{storeid});

    my $cmd = ['/usr/bin/qemu-img', 'convert', '-p', '-t', 'writeback', '-O', "$info->{target}->{format}", "$from", "$to"];
	$info->{source}->{plugin}->activate_volume($info->{source}->{storeid}, $info->{source}, $info->{source}->{volname});

	#print Dumper($cmd);
    run_command($cmd, errmsg => "migrate '$from->$to' error");	

    $info->{source}->{plugin}->deactivate_volume($info->{source}->{storeid}, $info->{source}, $info->{source}->{volname});
};

my $migrate_device_device = sub {
	my ($info) = @_;
	
	my $from_plugin = $info->{source}->{plugin};
	my $to_plugin = $info->{target}->{plugin};

	my $from = $from_plugin->path($info->{source}, $info->{source}->{volname}, $info->{source}->{storeid});
	my $to = $to_plugin->path($info->{target}, $info->{target}->{volname}, $info->{target}->{storeid});
    my $cmd = ['/usr/bin/qemu-img', 'convert', '-p', '-t', 'writeback', '-O', "$info->{target}->{format}", "$from", "$to"];
	$info->{source}->{plugin}->activate_volume($info->{source}->{storeid}, $info->{source}, $info->{source}->{volname});
	
	#print Dumper($cmd);
    run_command($cmd, errmsg => "migrate '$from->$to' error");
    
    $info->{source}->{plugin}->deactivate_volume($info->{source}->{storeid}, $info->{source}, $info->{source}->{volname});
};

my $update_config = sub {
	my ($info) = @_;
	my $source = $info->{source};
	my $target = $info->{target};
	my ($diskkey, $disk);
	
	my $format_store = sub {
		my ($src) = @_;
		
		return "$src->{storeid}:$src->{volname}";
	};
	
	my $updatefn =  sub {
		my ($disk, $newconf) = @_;

		my $conf = PVE::QemuServer::load_config($source->{vmid});

		PVE::QemuServer::check_lock($conf);
		
		if ($disk && $newconf) {
			my $line = $conf->{$disk};
			my @parts = split ',', $line;
			$parts[0] = $newconf;
			
			my $size = $target->{plugin}->volume_size_info($target, $target->{storeid}, $target->{volname});
			for (my $i = 0; $parts[$i]; $i++) {
				if ($parts[$i] =~ /^size=/) {
					my @s = split '=', $parts[$i];
					$s[1] = $size;
					$parts[$i] = join '=', @s;
					last;
				}
			}

			$line = join ',', @parts;
			$conf->{$disk} = $line;
			PVE::QemuServer::update_config_nolock($source->{vmid}, $conf, 1);
		}
	};
	
	PVE::QemuServer::lock_config($source->{vmid}, $updatefn);
	my $conf = PVE::QemuServer::load_config($source->{vmid});
	my $fmt = $format_store->($source);

	foreach my $key (disk_keys()) {
		if ($conf->{$key}) {
			$disk = $conf->{$key};
			if ($disk =~ /^\Q$fmt/) {
				$diskkey = $key;
				last;
			}
		}
	}

	#print Dumper($conf);
	die "update config failed. Storage $fmt not found for vm-$source->{vmid}.\n" unless $diskkey;
	PVE::QemuServer::lock_config($source->{vmid}, $updatefn, $diskkey, $format_store->($target));
};

my $remove_source = sub {
	my ($source) = @_;
	
	my $plugin = $source->{plugin};
	
	$unlock_source->($source);
	
	$plugin->free_image($source->{storeid}, $source, $source->{volname});
	#print Dumper($source);
};

# Factory is a wrapper unifying storage migration.
# Each supported storage migrate combination (from [format, storage], to [format, storage])
# will return a data structure (HASH) providing the following methods:
# create(migrate_info): create the target
# migrate(migrate_info): migrate source to target
# update(migrate_info): update VM configuration
# remove(migrate_info->source): remove the source
# Each method is a step which requires a successful prior step so each
# method must be called in the listed order.
# And the following information:
# type: type of migration
my $get_factory = sub {
	my ($info) = @_;
	my $factory = {};
	
	if ($info->{source}->{type} eq 'lvm' && $info->{target}->{type} eq 'lvm') {
		$factory->{type} = 'lvm2lvm';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_device_device;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} elsif ($info->{source}->{type} eq 'lvm' && $info->{target}->{type} eq 'nfs') {
		$factory->{type} = 'lvm2nfs';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_device_file;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} elsif ($info->{source}->{type} eq 'lvm' && $info->{target}->{type} eq 'dir') {
		$factory->{type} = 'lvm2dir';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_device_file;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} elsif ($info->{source}->{type} eq 'nfs' && $info->{target}->{type} eq 'nfs') {
		$factory->{type} = 'nfs2nfs';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_file_file;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} elsif ($info->{source}->{type} eq 'nfs' && $info->{target}->{type} eq 'lvm') {
		$factory->{type} = 'nfs2lvm';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_file_device;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} elsif ($info->{source}->{type} eq 'nfs' && $info->{target}->{type} eq 'dir') {
		$factory->{type} = 'nfs2dir';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_file_file;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} elsif ($info->{source}->{type} eq 'dir' && $info->{target}->{type} eq 'dir') {
		$factory->{type} = 'dir2dir';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_file_file;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} elsif ($info->{source}->{type} eq 'dir' && $info->{target}->{type} eq 'lvm') {
		$factory->{type} = 'dir2lvm';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_file_device;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} elsif ($info->{source}->{type} eq 'dir' && $info->{target}->{type} eq 'nfs') {
		$factory->{type} = 'dir2nfs';
		$factory->{create} = \&$create_target;
		$factory->{migrate} = \&$migrate_file_file;
		$factory->{update} = \&$update_config;
		$factory->{remove} = \&$remove_source;
		return $factory;
	} else {
		die $info->{source}->{type} . "->" . $info->{target}->{type} . ": Not implemented\n";
	}
};

my $execute = sub {
	my ($info, $factory) = @_;
	my $err;

	my $remove_target = sub {
		my ($target) = @_;
		
		my $plugin = $target->{plugin};
		$plugin->free_image($target->{storeid}, $target, $target->{volname});
	};
	
	my $error = sub {
		my ($phase, $err) = @_;
		my %need_remove_target = (
			'migrate' => 1,
			'update' => 1,
			'remove' => 1,
		);
		
		if ($need_remove_target{$phase}) {
			eval {
				$remove_target->($info->{target});
			};
			print $@ if $@;
		}
		eval {
			$unlock_source->($info->{source});
		};
		print $@ if $@;
		die $err;
	};
	
	print "Storage migrate type: $factory->{type}\n";
	print "Create target...\n";
	eval {
		$factory->{create}->($info);
	};
	$error->('create', $@) if $@;
	
	print "Migrate source...\n";
	eval {
		$factory->{migrate}->($info);
	};
	$error->('migrate', $@) if $@;

	print "Update vm config...\n";
	eval {
		$factory->{update}->($info);
	};
	$error->('update', $@) if $@;
	
	print "Remove source...\n";
	eval {
		$factory->{remove}->($info->{source});
	};
	$error->('remove', $@) if $@;

#	my $plugin = $info->{target}->{plugin};
#	$plugin->free_image($info->{target}->{storeid}, $info->{target}, $info->{target}->{volname});
};

sub storage_migrate {
	my ($volid, $target) = @_;

	my $cfg = PVE::Storage::config();
    my ($storeid, $volname) = PVE::Storage::parse_volume_id($volid, 1);

	if (PVE::QemuServer::check_running(vm_name($volname))) {
		# Only migrate if VM is off-line
		die "vm-" . vm_name($volname) . ": Running - Only offline storage migration supported.\n";
	}
	
	my $info = $migrate_info->($cfg, $storeid, $volname, $target);

	# if volid belongs to a HA enabled VM target must be shared
	if (volume_ha_enabled($volname)) {
		unless ($info->{target}->{shared}) {
			die vm_name($volname) . ": HA and target is not shared\n";
		}
	}
	#print Dumper($info);
	
	my $factory = $get_factory->($info);

	$execute->($info, $factory);
}

#storage_migrate('pve-storage1_lvm:vm-102-disk-1', 'pve-storage1_nfs');
#storage_migrate('pve-storage1_nfs:102/vm-102-disk-2.qcow2', 'pve-storage2_lvm');
#storage_migrate('pve-storage2_lvm:vm-102-disk-2', 'pve-storage1_nfs');
#storage_migrate('pve-storage2_nfs:100/vm-100-disk-1.vmdk', 'pve-storage1_nfs');
#storage_migrate('pve-storage1_nfs:102/vm-102-disk-2.qcow2', 'pve-storage2_nfs');

#print Dumper(PVE::Storage::Plugin->path(PVE::Storage::storage_config(PVE::Storage::config(), 'pve-storage2_lvm'), 'vm-100-disk1'));
1;

